@@ -32,7 +32,6 @@
 #include "archiveformat.h"
 #include "ark_debug.h"
 #include "ui_createdialog.h"
-#include "kerfuffle/archive_kerfuffle.h"
 #include "mimetypes.h"
 
 #include <KMessageBox>
@@ -50,7 +49,7 @@ class CreateDialogUI: public QWidget, public Ui::CreateDialog
     Q_OBJECT
 
 public:
-    CreateDialogUI(QWidget *parent = 0)
+    CreateDialogUI(QWidget *parent = nullptr)
             : QWidget(parent) {
         setupUi(this);
     }
@@ -90,8 +89,8 @@ CreateDialog::CreateDialog(QWidget *parent,
     connect(m_ui->buttonBox, &QDialogButtonBox::accepted, this, &QDialog::accept);
     connect(m_ui->buttonBox, &QDialogButtonBox::rejected, this, &QDialog::reject);
     connect(this, &QDialog::accepted, this, &CreateDialog::slotUpdateDefaultMimeType);
-    connect(m_ui->mimeComboBox, static_cast<void(QComboBox::*)(int)>(&QComboBox::currentIndexChanged), this, &CreateDialog::slotUpdateWidgets);
-    connect(m_ui->mimeComboBox, static_cast<void(QComboBox::*)(int)>(&QComboBox::currentIndexChanged), this, &CreateDialog::slotUpdateFilenameExtension);
+    connect(m_ui->mimeComboBox, QOverload<int>::of(&QComboBox::currentIndexChanged), this, &CreateDialog::slotUpdateWidgets);
+    connect(m_ui->mimeComboBox, QOverload<int>::of(&QComboBox::currentIndexChanged), this, &CreateDialog::slotUpdateFilenameExtension);
 
     m_vlayout->addWidget(m_ui);
 
@@ -206,7 +205,7 @@ void CreateDialog::accept()
         QDialog::accept();
         break;
     case KNewPasswordWidget::PasswordNotVerified:
-        KMessageBox::error(Q_NULLPTR, i18n("The chosen password does not match the given verification password."));
+        KMessageBox::error(nullptr, i18n("The chosen password does not match the given verification password."));
         break;
     default:
         break;
@@ -233,7 +232,6 @@ QMimeType CreateDialog::currentMimeType() const
 
 bool CreateDialog::setMimeType(const QString &mimeTypeName)
 {
-    const QMimeType mimeType = QMimeDatabase().mimeTypeForName(mimeTypeName);
     int index = m_supportedMimeTypes.indexOf(mimeTypeName);
     if (index == -1) {
         return false;
