@@ -36,9 +36,11 @@
 #include <QDir>
 #include <QDirIterator>
 #include <QFile>
+#include <qplatformdefs.h>
 #include <QThread>
 
 #include <utime.h>
+#include <zlib.h>
 
 K_PLUGIN_FACTORY_WITH_JSON(LibZipPluginFactory, "kerfuffle_libzip.json", registerPlugin<LibzipPlugin>();)
 
@@ -95,7 +97,7 @@ bool LibzipPlugin::list()
     m_comment = QString::fromUtf8(zip_get_archive_comment(archive, nullptr, ZIP_FL_ENC_GUESS));
 
     // Get number of archive entries.
-    int nofEntries = zip_get_num_entries(archive, 0);
+    const auto nofEntries = zip_get_num_entries(archive, 0);
     qCDebug(ARK) << "Found entries:" << nofEntries;
 
     // Loop through all archive entries.
@@ -155,7 +157,7 @@ bool LibzipPlugin::addFiles(const QVector<Archive::Entry*> &files, const Archive
                             QDirIterator::Subdirectories);
 
             while (!QThread::currentThread()->isInterruptionRequested() && it.hasNext()) {
-                QString path = it.next();
+                const QString path = it.next();
 
                 if (QFileInfo(path).isDir()) {
                     if (!writeEntry(archive, path, destination, options, true)) {
@@ -234,6 +236,20 @@ bool LibzipPlugin::writeEntry(zip_t *archive, const QString &file, const Archive
             return false;
         }
     }
+
+#ifndef Q_OS_WIN
+    // Set permissions.
+    QT_STATBUF result;
+    if (QT_STAT(QFile::encodeName(file), &result) != 0) {
+        qCWarning(ARK) << "Failed to read permissions for:" << file;
+    } else {
+        zip_uint32_t attributes = result.st_mode << 16;
+        if (zip_file_set_external_attributes(archive, index, ZIP_FL_UNCHANGED, ZIP_OPSYS_UNIX, attributes) != 0) {
+            qCWarning(ARK) << "Failed to set external attributes for:" << file;
+        }
+    }
+#endif
+
     if (!password().isEmpty()) {
         Q_ASSERT(!options.encryptionMethod().isEmpty());
         if (options.encryptionMethod() == QLatin1String("AES128")) {
@@ -245,6 +261,24 @@ bool LibzipPlugin::writeEntry(zip_t *archive, const QString &file, const Archive
         }
     }
 
+    // Set compression level and method.
+    zip_int32_t compMethod = ZIP_CM_DEFAULT;
+    if (!options.compressionMethod().isEmpty()) {
+        if (options.compressionMethod() == QLatin1String("Deflate")) {
+            compMethod = ZIP_CM_DEFLATE;
+        } else if (options.compressionMethod() == QLatin1String("BZip2")) {
+            compMethod = ZIP_CM_BZIP2;
+        } else if (options.compressionMethod() == QLatin1String("Store")) {
+            compMethod = ZIP_CM_STORE;
+        }
+    }
+    const int compLevel = options.isCompressionLevelSet() ? options.compressionLevel() : 6;
+    if (zip_set_file_compression(archive, index, compMethod, compLevel) != 0) {
+        qCCritical(ARK) << "Could not set compression options for" << file << ":" << zip_strerror(archive);
+        emit error(xi18n("Failed to set compression options for entry: %1", QString::fromUtf8(zip_strerror(archive))));
+        return false;
+    }
+
     return true;
 }
 
@@ -252,38 +286,38 @@ bool LibzipPlugin::emitEntryForIndex(zip_t *archive, qlonglong index)
 {
     Q_ASSERT(archive);
 
-    zip_stat_t sb;
-    if (zip_stat_index(archive, index, ZIP_FL_ENC_GUESS, &sb)) {
+    zip_stat_t statBuffer;
+    if (zip_stat_index(archive, index, ZIP_FL_ENC_GUESS, &statBuffer)) {
         qCCritical(ARK) << "Failed to read stat for index" << index;
         return false;
     }
 
     auto e = new Archive::Entry();
 
-    if (sb.valid & ZIP_STAT_NAME) {
-        e->setFullPath(QString::fromUtf8(sb.name));
+    if (statBuffer.valid & ZIP_STAT_NAME) {
+        e->setFullPath(QString::fromUtf8(statBuffer.name));
     }
 
     if (e->fullPath(PathFormat::WithTrailingSlash).endsWith(QDir::separator())) {
         e->setProperty("isDirectory", true);
     }
 
-    if (sb.valid & ZIP_STAT_MTIME) {
-        e->setProperty("timestamp", QDateTime::fromTime_t(sb.mtime));
+    if (statBuffer.valid & ZIP_STAT_MTIME) {
+        e->setProperty("timestamp", QDateTime::fromTime_t(statBuffer.mtime));
     }
-    if (sb.valid & ZIP_STAT_SIZE) {
-        e->setProperty("size", (qulonglong)sb.size);
+    if (statBuffer.valid & ZIP_STAT_SIZE) {
+        e->setProperty("size", (qulonglong)statBuffer.size);
     }
-    if (sb.valid & ZIP_STAT_COMP_SIZE) {
-        e->setProperty("compressedSize", (qlonglong)sb.comp_size);
+    if (statBuffer.valid & ZIP_STAT_COMP_SIZE) {
+        e->setProperty("compressedSize", (qlonglong)statBuffer.comp_size);
     }
-    if (sb.valid & ZIP_STAT_CRC) {
+    if (statBuffer.valid & ZIP_STAT_CRC) {
         if (!e->isDir()) {
-            e->setProperty("CRC", QString::number((qulonglong)sb.crc, 16).toUpper());
+            e->setProperty("CRC", QString::number((qulonglong)statBuffer.crc, 16).toUpper());
         }
     }
-    if (sb.valid & ZIP_STAT_COMP_METHOD) {
-        switch(sb.comp_method) {
+    if (statBuffer.valid & ZIP_STAT_COMP_METHOD) {
+        switch(statBuffer.comp_method) {
             case ZIP_CM_STORE:
                 e->setProperty("method", QStringLiteral("Store"));
                 emit compressionMethodFound(QStringLiteral("Store"));
@@ -310,10 +344,10 @@ bool LibzipPlugin::emitEntryForIndex(zip_t *archive, qlonglong index)
                 break;
         }
     }
-    if (sb.valid & ZIP_STAT_ENCRYPTION_METHOD) {
-        if (sb.encryption_method != ZIP_EM_NONE) {
+    if (statBuffer.valid & ZIP_STAT_ENCRYPTION_METHOD) {
+        if (statBuffer.encryption_method != ZIP_EM_NONE) {
             e->setProperty("isPasswordProtected", true);
-            switch(sb.encryption_method) {
+            switch(statBuffer.encryption_method) {
                 case ZIP_EM_TRAD_PKWARE:
                     emit encryptionMethodFound(QStringLiteral("ZipCrypto"));
                     break;
@@ -330,6 +364,25 @@ bool LibzipPlugin::emitEntryForIndex(zip_t *archive, qlonglong index)
         }
     }
 
+    // Read external attributes, which contains the file permissions.
+    zip_uint8_t opsys;
+    zip_uint32_t attributes;
+    if (zip_file_get_external_attributes(archive, index, ZIP_FL_UNCHANGED, &opsys, &attributes) == -1) {
+        qCCritical(ARK) << "Could not read external attributes for entry:" << QString::fromUtf8(statBuffer.name);
+        emit error(xi18n("Failed to read metadata for entry: %1", QString::fromUtf8(statBuffer.name)));
+        return false;
+    }
+
+    // Set permissions.
+    switch (opsys) {
+    case ZIP_OPSYS_UNIX:
+        // Unix permissions are stored in the leftmost 16 bits of the external file attribute.
+        e->setProperty("permissions", permissionsToString(attributes >> 16));
+        break;
+    default:    // TODO: non-UNIX.
+        break;
+    }
+
     emit entry(e);
     m_emittedEntries << e;
 
@@ -357,7 +410,7 @@ bool LibzipPlugin::deleteFiles(const QVector<Archive::Entry*> &files)
             break;
         }
 
-        qlonglong index = zip_name_locate(archive, e->fullPath().toUtf8(), ZIP_FL_ENC_GUESS);
+        const qlonglong index = zip_name_locate(archive, e->fullPath().toUtf8(), ZIP_FL_ENC_GUESS);
         if (index == -1) {
             qCCritical(ARK) << "Could not find entry to delete:" << e->fullPath();
             emit error(xi18n("Failed to delete entry: %1", e->fullPath()));
@@ -423,6 +476,33 @@ bool LibzipPlugin::testArchive()
         qCCritical(ARK) << "Failed to open archive:" << zip_error_strerror(&err);
         return false;
     }
+
+    // Check CRC-32 for each archive entry.
+    const int nofEntries = zip_get_num_entries(archive, 0);
+    for (int i = 0; i < nofEntries; i++) {
+
+        // Get statistic for entry. Used to get entry size.
+        zip_stat_t statBuffer;
+        if (zip_stat_index(archive, i, 0, &statBuffer) != 0) {
+            qCCritical(ARK) << "Failed to read stat for" << statBuffer.name;
+            return false;
+        }
+
+        zip_file *zipFile = zip_fopen_index(archive, i, 0);
+        QScopedPointer<uchar> buf(new uchar[statBuffer.size]);
+        const int len = zip_fread(zipFile, buf.data(), statBuffer.size);
+        if (len == -1 || uint(len) != statBuffer.size) {
+            qCCritical(ARK) << "Failed to read data for" << statBuffer.name;
+            return false;
+        }
+        if (statBuffer.crc != crc32(0, &buf.data()[0], len)) {
+            qCCritical(ARK) << "CRC check failed for" << statBuffer.name;
+            return false;
+        }
+
+        emit progress(float(i) / nofEntries);
+    }
+
     zip_close(archive);
 
     emit testSuccess();
@@ -461,8 +541,7 @@ bool LibzipPlugin::extractFiles(const QVector<Archive::Entry*> &files, const QSt
     }
 
     // Get number of archive entries.
-    qlonglong nofEntries;
-    extractAll ? nofEntries = zip_get_num_entries(archive, 0) : nofEntries = files.size();
+    const qlonglong nofEntries = extractAll ? zip_get_num_entries(archive, 0) : files.size();
 
     // Extract entries.
     m_overwriteAll = false; // Whether to overwrite all files
@@ -545,8 +624,8 @@ bool LibzipPlugin::extractEntry(zip_t *archive, const QString &entry, const QStr
     } else {
         parentDir = QFileInfo(destination).path();
     }
-    // For top-level items, dont restore parent dir mtime.
-    bool restoreParentMtime = (parentDir + QDir::separator() != destDirCorrected);
+    // For top-level items, don't restore parent dir mtime.
+    const bool restoreParentMtime = (parentDir + QDir::separator() != destDirCorrected);
 
     time_t parent_mtime;
     if (restoreParentMtime) {
@@ -561,8 +640,8 @@ bool LibzipPlugin::extractEntry(zip_t *archive, const QString &entry, const QStr
     }
 
     // Get statistic for entry. Used to get entry size and mtime.
-    zip_stat_t sb;
-    if (zip_stat(archive, entry.toUtf8(), 0, &sb) != 0) {
+    zip_stat_t statBuffer;
+    if (zip_stat(archive, entry.toUtf8(), 0, &statBuffer) != 0) {
         qCCritical(ARK) << "Failed to read stat for entry" << entry;
         return false;
     }
@@ -600,11 +679,11 @@ bool LibzipPlugin::extractEntry(zip_t *archive, const QString &entry, const QStr
         }
 
         // Handle password-protected files.
-        zip_file *zf = nullptr;
+        zip_file *zipFile = nullptr;
         bool firstTry = true;
-        while (!zf) {
-            zf = zip_fopen(archive, entry.toUtf8(), 0);
-            if (zf) {
+        while (!zipFile) {
+            zipFile = zip_fopen(archive, entry.toUtf8(), 0);
+            if (zipFile) {
                 break;
             } else if (zip_error_code_zip(zip_get_error(archive)) == ZIP_ER_NOPASSWD ||
                        zip_error_code_zip(zip_get_error(archive)) == ZIP_ER_WRONGPASSWD) {
@@ -641,21 +720,20 @@ bool LibzipPlugin::extractEntry(zip_t *archive, const QString &entry, const QStr
         // Write archive entry to file. We use a read/write buffer of 1000 chars.
         qulonglong sum = 0;
         char buf[1000];
-        int len;
-        while (sum != sb.size) {
-            len = zip_fread(zf, buf, 1000);
-            if (len < 0) {
+        while (sum != statBuffer.size) {
+            const auto readBytes = zip_fread(zipFile, buf, 1000);
+            if (readBytes < 0) {
                 qCCritical(ARK) << "Failed to read data";
                 emit error(xi18n("Failed to read data for entry: %1", entry));
                 return false;
             }
-            if (out.writeRawData(buf, len) != len) {
+            if (out.writeRawData(buf, readBytes) != readBytes) {
                 qCCritical(ARK) << "Failed to write data";
                 emit error(xi18n("Failed to write data for entry: %1", entry));
                 return false;
             }
 
-            sum += len;
+            sum += readBytes;
         }
 
         const auto index = zip_name_locate(archive, entry.toUtf8(), ZIP_FL_ENC_GUESS);
@@ -688,7 +766,7 @@ bool LibzipPlugin::extractEntry(zip_t *archive, const QString &entry, const QStr
 
     // Set mtime for entry.
     utimbuf times;
-    times.modtime = sb.mtime;
+    times.modtime = statBuffer.mtime;
     if (utime(destination.toUtf8(), &times) != 0) {
         qCWarning(ARK) << "Failed to restore mtime:" << destination;
     }
@@ -807,17 +885,87 @@ bool LibzipPlugin::copyFiles(const QVector<Archive::Entry*> &files, Archive::Ent
             return false;
         }
 
-        emitEntryForIndex(archive, destIndex);
-        emit progress(i/filePaths.count());
+        // Get permissions from source entry.
+        zip_uint8_t opsys;
+        zip_uint32_t attributes;
+        if (zip_file_get_external_attributes(archive, srcIndex, ZIP_FL_UNCHANGED, &opsys, &attributes) == -1) {
+            qCCritical(ARK) << "Failed to read external attributes for source:" << filePaths.at(i);
+            emit error(xi18n("Failed to read metadata for entry: %1", filePaths.at(i)));
+            return false;
+        }
+
+        // Set permissions on dest entry.
+        if (zip_file_set_external_attributes(archive, destIndex, ZIP_FL_UNCHANGED, opsys, attributes) != 0) {
+            qCCritical(ARK) << "Failed to set external attributes for destination:" << dest;
+            emit error(xi18n("Failed to set metadata for entry: %1", dest));
+            return false;
+        }
     }
+
+    // Register the callback function to get progress feedback.
+    Callback<void(double)>::func = std::bind(&LibzipPlugin::progressEmitted, this, std::placeholders::_1);
+    void (*c_func)(double) = static_cast<decltype(c_func)>(Callback<void(double)>::callback);
+    zip_register_progress_callback(archive, c_func);
+
     if (zip_close(archive)) {
         qCCritical(ARK) << "Failed to write archive";
         emit error(xi18n("Failed to write archive."));
         return false;
     }
+
+    // List the archive to update the model.
+    m_listAfterAdd = true;
+    list();
+
     qCDebug(ARK) << "Copied" << i << "entries";
 
     return true;
 }
 
+QString LibzipPlugin::permissionsToString(const mode_t &perm)
+{
+    QString modeval;
+    if ((perm & S_IFMT) == S_IFDIR) {
+        modeval.append(QLatin1Char('d'));
+    } else if ((perm & S_IFMT) == S_IFLNK) {
+        modeval.append(QLatin1Char('l'));
+    } else {
+        modeval.append(QLatin1Char('-'));
+    }
+    modeval.append((perm & S_IRUSR) ? QLatin1Char('r') : QLatin1Char('-'));
+    modeval.append((perm & S_IWUSR) ? QLatin1Char('w') : QLatin1Char('-'));
+    if ((perm & S_ISUID) && (perm & S_IXUSR)) {
+        modeval.append(QLatin1Char('s'));
+    } else if ((perm & S_ISUID)) {
+        modeval.append(QLatin1Char('S'));
+    } else if ((perm & S_IXUSR)) {
+        modeval.append(QLatin1Char('x'));
+    } else {
+        modeval.append(QLatin1Char('-'));
+    }
+    modeval.append((perm & S_IRGRP) ? QLatin1Char('r') : QLatin1Char('-'));
+    modeval.append((perm & S_IWGRP) ? QLatin1Char('w') : QLatin1Char('-'));
+    if ((perm & S_ISGID) && (perm & S_IXGRP)) {
+        modeval.append(QLatin1Char('s'));
+    } else if ((perm & S_ISGID)) {
+        modeval.append(QLatin1Char('S'));
+    } else if ((perm & S_IXGRP)) {
+        modeval.append(QLatin1Char('x'));
+    } else {
+        modeval.append(QLatin1Char('-'));
+    }
+    modeval.append((perm & S_IROTH) ? QLatin1Char('r') : QLatin1Char('-'));
+    modeval.append((perm & S_IWOTH) ? QLatin1Char('w') : QLatin1Char('-'));
+    if ((perm & S_ISVTX) && (perm & S_IXOTH)) {
+        modeval.append(QLatin1Char('t'));
+    } else if ((perm & S_ISVTX)) {
+        modeval.append(QLatin1Char('T'));
+    } else if ((perm & S_IXOTH)) {
+        modeval.append(QLatin1Char('x'));
+    } else {
+        modeval.append(QLatin1Char('-'));
+    }
+    return modeval;
+}
+
 #include "libzipplugin.moc"
